# Copyright (c) 2019 SUSE Linux GmbH.  All rights reserved.
#
# This file is part of kiwi.
#
# kiwi is free software: you can redistribute it and/or modify
# it under the terms owf the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# kiwi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kiwi.  If not, see <http://www.gnu.org/licenses/>
#
import logging
import os
import random
import string

# project
from kiwi.oci_tools.base import OCIBase
from kiwi.command import Command
from kiwi.path import Path
from kiwi.defaults import Defaults
from kiwi.exceptions import KiwiBuildahError

log = logging.getLogger('kiwi')


class OCIBuildah(OCIBase):
    """
    **Open Container Operations using buildah**
    """
    def post_init(self):
        """
        Initializes some default parameters
        """
        self.working_image = None
        self.imported_image = None
        self.working_container = None

    def __enter__(self):
        return self

    def import_container_image(self, container_image_ref):
        """
        Imports container image reference to the OCI containers storage.

        :param str container_image_ref: container image reference
        """
        if not self.imported_image:
            self.imported_image = 'kiwi-image-{0}:{1}'.format(
                self._random_string_generator(),
                Defaults.get_container_base_image_tag()
            )
        else:
            raise KiwiBuildahError(
                "Image already imported, called: '{0}'".format(
                    self.imported_image
                )
            )

        # We are making use of skopeo instead of only calling 'buildah from'
        # because we want to control the image name loaded into the containers
        # storage. This way we are certain to not leave any left over after the
        # build.
        Command.run(
            [
                'skopeo', 'copy', container_image_ref,
                'containers-storage:{0}'.format(self.imported_image)
            ] + (
                [
                    '--tmpdir', Defaults.get_temp_location()
                ] if self._skopeo_provides_tmpdir_option() else []
            )
        )

        if not self.working_container:
            self.working_container = 'kiwi-container-{0}'.format(
                self._random_string_generator()
            )
        else:
            raise KiwiBuildahError(
                "Container already initated, called: '{0}'".format(
                    self.working_container
                )
            )

        Command.run(
            [
                'buildah', 'from', '--name', self.working_container,
                'containers-storage:{0}'.format(self.imported_image)
            ]
        )

    def export_container_image(
        self, filename, transport, image_ref, additional_names=None
    ):
        """
        Exports the working container to a container image archive

        :param str filename: The resulting filename
        :param str transport: The archive format
        :param str image_ref: Reference of the exported image
        :param list additional_names: List of additional references
        """
        extra_tags_opt = []
        if additional_names:
            for ref in additional_names:
                extra_tags_opt.extend(['--additional-tag', ref])

        # make sure the target tar file does not exist
        # skopeo doesn't support force overwrite
        Path.wipe(filename)
        if self.working_image:
            export_image = self.working_image
        elif self.imported_image:
            export_image = self.imported_image
        else:
            raise KiwiBuildahError("There is no image to export defined")

        # we are using 'skopeo copy' to export images instead of 'buildah push'
        # because buildah does not support multiple tags
        Command.run(
            [
                'skopeo', 'copy', 'containers-storage:{0}'.format(export_image),
                '{0}:{1}:{2}'.format(transport, filename, image_ref)
            ] + extra_tags_opt + (
                [
                    '--tmpdir', Defaults.get_temp_location()
                ] if self._skopeo_provides_tmpdir_option() else []
            )
        )

    def init_container(self):
        """
        Initialize a new container in OCI containers storage
        """
        if not self.working_container:
            self.working_container = 'kiwi-container-{0}'.format(
                self._random_string_generator()
            )
        else:
            raise KiwiBuildahError(
                "Image already imported or initated at '{0}' container".format(
                    self.working_container
                )
            )

        Command.run(
            ['buildah', 'from', '--name', self.working_container, 'scratch']
        )

    def unpack(self):
        """
        Mounts current container root data to a directory
        """
        cmd = Command.run(
            ['buildah', 'mount', self.working_container]
        )
        self.oci_root_dir = cmd.output.rstrip()

    def sync_rootfs(self, root_dir, exclude_list=None):
        """
        Synchronizes the image root with the rootfs of the container

        :param string root_dir: root directory of the prepare step
        :param list exclude_list: list of paths to exclude
        """
        self._sync_data(
            ''.join([root_dir, os.sep]), self.oci_root_dir,
            exclude_list=exclude_list,
            options=Defaults.get_sync_options() + [
                '--filter', '-x! user.*',
                '--filter', '-x! security.ima*',
                '--filter', '-x! security.capability*',
                '--delete'
            ]
        )

    def import_rootfs(self, root_dir, exclude_list=None):
        """
        Synchronizes the container rootfs with the root tree of the build

        :param string root_dir: root directory used in prepare step
        :param list exclude_list: list of paths to exclude
        """
        self._sync_data(
            os.sep.join([self.oci_root_dir, '']), root_dir,
            exclude_list=exclude_list,
            options=Defaults.get_sync_options()
        )

    def repack(self, oci_config):
        """
        Pack root data directory into container image

        :param list oci_config: unused parameter
        """
        Command.run(
            ['buildah', 'umount', self.working_container]
        )

    def set_config(self, oci_config):
        """
        Set list of meta data information such as entry_point,
        maintainer, etc... to the container.

        :param list oci_config: meta data list
        :param bool base_image: True|False
        """
        config_args = self._process_oci_config_to_arguments(oci_config)
        Command.run(
            ['buildah', 'config'] + config_args + [self.working_container]
        )

    def post_process(self):
        """
        Commits the OCI container into an OCI image
        """
        if not self.working_image and self.working_container:
            self.working_image = 'kiwi-image-{0}:{1}'.format(
                self._random_string_generator(), 'tag-{0}'.format(
                    self._random_string_generator()
                )
            )
        else:
            raise KiwiBuildahError(
                "No container to commit or container already committed"
            )

        output = Command.run(
            [
                'buildah', 'commit', '--rm', '--format', 'oci',
                self.working_container, self.working_image
            ]
        )
        self.working_image = output.output.rstrip()
        self.working_container = None

    @classmethod
    def _process_oci_config_to_arguments(self, oci_config):
        """
        Process the oci configuration dictionary into a list of arguments
        for the 'buildah config' command

        :param list oci_config: meta data list

        :return: List of buildah config arguments

        :rtype: list
        """
        arguments = []
        if 'maintainer' in oci_config:
            arguments.append(
                '--author={0}'.format(oci_config['maintainer'])
            )

        if 'user' in oci_config:
            arguments.append(
                '--user={0}'.format(oci_config['user'])
            )

        if 'workingdir' in oci_config:
            arguments.append(
                '--workingdir={0}'.format(oci_config['workingdir'])
            )

        if 'entry_command' in oci_config:
            arguments.append('--entrypoint=[{0}]'.format(
                ','.join(
                    ['"{0}"'.format(x) for x in oci_config['entry_command']]
                )
            ))

        if 'entry_subcommand' in oci_config:
            arguments.append('--cmd={0}'.format(
                ' '.join(oci_config['entry_subcommand'])
            ))

        if 'volumes' in oci_config:
            for vol in oci_config['volumes']:
                arguments.append('--volume={0}'.format(vol))

        if 'stopsignal' in oci_config:
            arguments.append(
                '--stop-signal={0}'.format(oci_config['stopsignal'])
            )

        if 'expose_ports' in oci_config:
            for port in oci_config['expose_ports']:
                arguments.append('--port={0}'.format(port))

        if 'environment' in oci_config:
            for name in sorted(oci_config['environment']):
                arguments.append('--env={0}={1}'.format(
                    name, oci_config['environment'][name]
                ))

        if 'labels' in oci_config:
            for name in sorted(oci_config['labels']):
                arguments.append('--label={0}={1}'.format(
                    name, oci_config['labels'][name]
                ))

        if 'history' in oci_config:
            if 'comment' in oci_config['history']:
                arguments.append('--history-comment={0}'.format(
                    oci_config['history']['comment']
                ))
            if 'created_by' in oci_config['history']:
                arguments.append('--created-by={0}'.format(
                    oci_config['history']['created_by']
                ))
            if 'author' in oci_config['history']:
                log.warning('Author field in history is ignored using buildah')

        return arguments

    @classmethod
    def _random_string_generator(
        cls, chars_num=6, allchars=string.ascii_lowercase + string.digits
    ):
        """
        Creates a random string with the given length of characters choosen
        randomly from a given list of possible characters.

        Buildah makes use of the hosts configured containers storage of OCI
        images. This method is used to avoid name collisions with any
        previous image or container present in the host.

        :param int chars_num: Lenght of the generated random string
        :param list allchars: List of possible characters

        :return: generated random string

        :rtype: string
        """
        return "".join(random.choice(allchars) for x in range(chars_num))

    def __exit__(self, exc_type, exc_value, traceback):
        if self.working_container:
            Command.run(['buildah', 'umount', self.working_container])
            Command.run(['buildah', 'rm', self.working_container])
        if self.working_image:
            Command.run(['buildah', 'rmi', self.working_image])
        if self.imported_image:
            Command.run(['buildah', 'rmi', self.imported_image])



from unittest.mock import patch
from pytest import raises

from kiwi.container.setup import ContainerSetup

from kiwi.exceptions import KiwiContainerSetupError


class TestContainerSetup:
    def test_container_not_implemented(self):
        with raises(KiwiContainerSetupError):
            ContainerSetup.new('foo', 'root_dir')

    @patch('kiwi.container.setup.docker.ContainerSetupDocker')
    def test_container_docker(self, mock_docker):
        ContainerSetup.new('docker', 'root_dir')
        mock_docker.assert_called_once_with('root_dir', None)

    @patch('kiwi.container.setup.oci.ContainerSetupOCI')
    def test_container_oci(self, mock_oci):
        ContainerSetup.new('oci', 'root_dir')
        mock_oci.assert_called_once_with('root_dir', None)

    @patch('kiwi.container.setup.appx.ContainerSetupAppx')
    def test_container_appx(self, mock_appx):
        ContainerSetup.new('appx', 'root_dir')
        mock_appx.assert_called_once_with('root_dir', None)



# Copyright (c) 2015 SUSE Linux GmbH.  All rights reserved.
#
# This file is part of kiwi.
#
# kiwi is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# kiwi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kiwi.  If not, see <http://www.gnu.org/licenses/>
#
"""
usage: kiwi-ng system build -h | --help
       kiwi-ng system build --description=<directory> --target-dir=<directory>
           [--allow-existing-root]
           [--clear-cache]
           [--ignore-repos]
           [--ignore-repos-used-for-build]
           [--set-repo=<source,type,alias,priority,imageinclude,package_gpgcheck,{signing_keys},components,distribution,repo_gpgcheck,repo_sourcetype>]
           [--set-repo-credentials=<user:pass_or_filename>]
           [--add-repo=<source,type,alias,priority,imageinclude,package_gpgcheck,{signing_keys},components,distribution,repo_gpgcheck,repo_sourcetype>...]
           [--add-repo-credentials=<user:pass_or_filename>...]
           [--add-package=<name>...]
           [--add-bootstrap-package=<name>...]
           [--delete-package=<name>...]
           [--set-container-derived-from=<uri>]
           [--set-container-tag=<name>]
           [--add-container-label=<label>...]
           [--set-type-attr=<attribute=value>...]
           [--set-release-version=<version>]
           [--signing-key=<key-file>...]
       kiwi-ng system build help

commands:
    build
        build a system image from the specified description. The
        build command combines the prepare and create commands
    build help
        show manual page for build command

options:
    --add-bootstrap-package=<name>
        install the given package name as part of the early bootstrap process
    --add-package=<name>
        install the given package name
    --add-repo=<source,type,alias,priority,imageinclude,package_gpgcheck,{signing_keys},components,distribution,repo_gpgcheck,repo_sourcetype>
        add repository with given source, type, alias,
        priority, imageinclude(true|false), package_gpgcheck(true|false),
        list of signing_keys enclosed in curly brackets delimited by a colon,
        component list for debian based repos as string delimited by a space,
        main distribution name for debian based repos,
        repo_gpgcheck(true|false) and repo_sourcetype(metalink|baseurl|mirrorlist)
    --add-repo-credentials=<user:pass_or_filename>
        for uri://user:pass@location type repositories, set the user and
        password connected with an add-repo specification. The first
        add-repo-credentials is connected with the first add-repo
        specification and so on. If the provided value describes a
        filename in the filesystem, the first line of that file is read
        and used as credentials information.
    --allow-existing-root
        allow to use an existing root directory from an earlier
        build attempt. Use with caution this could cause an inconsistent
        root tree if the existing contents does not fit to the
        former image type setup
    --clear-cache
        delete repository cache for each of the used repositories
        before installing any package
    --delete-package=<name>
        delete the given package name
    --description=<directory>
        the description must be a directory containing a kiwi XML
        description and optional metadata files
    --ignore-repos
        ignore all repos from the XML configuration
    --ignore-repos-used-for-build
        ignore all repos from the XML configuration except the
        ones marked as imageonly
    --set-container-derived-from=<uri>
        overwrite the source location of the base container
        for the selected image type. The setting is only effective
        if the configured image type is setup with an initial
        derived_from reference
    --set-container-tag=<name>
        overwrite the container tag in the container configuration.
        The setting is only effective if the container configuraiton
        provides an initial tag value
    --add-container-label=<name=value>
        add a container label in the container configuration metadata. It
        overwrites the label with the provided key-value pair in case it was
        already defined in the XML description
    --set-repo=<source,type,alias,priority,imageinclude,package_gpgcheck,{signing_keys},components,distribution,repo_gpgcheck,repo_sourcetype>
        overwrite the first XML listed repository source, type, alias,
        priority, imageinclude(true|false), package_gpgcheck(true|false),
        list of signing_keys enclosed in curly brackets delimited by a colon,
        component list for debian based repos as string delimited by a space,
        main distribution name for debian based repos,
        repo_gpgcheck(true|false) and repo_sourcetype(metalink|baseurl|mirrorlist)
    --set-repo-credentials=<user:pass_or_filename>
        for uri://user:pass@location type repositories, set the user and
        password connected to the set-repo specification. If the provided
        value describes a filename in the filesystem, the first line of that
        file is read and used as credentials information.
    --set-type-attr=<attribute=value>
        overwrite/set the attribute with the provided value in the selected
        build type section
    --set-release-version=<version>
        overwrite/set the release-version element in the selected
        build type preferences section
    --signing-key=<key-file>
        includes the key-file as a trusted key for package manager validations
    --target-dir=<directory>
        the target directory to store the system image file(s)
"""
import os
import logging
from itertools import zip_longest
from urllib.parse import urlparse

# project
from kiwi.tasks.base import CliTask
from kiwi.help import Help
from kiwi.system.prepare import SystemPrepare
from kiwi.system.setup import SystemSetup
from kiwi.builder import ImageBuilder
from kiwi.system.profile import Profile
from kiwi.defaults import Defaults
from kiwi.privileges import Privileges
from kiwi.path import Path

log = logging.getLogger('kiwi')


class SystemBuildTask(CliTask):
    """
    Implements building of system images

    Attributes

    * :attr:`manual`
        Instance of Help
    """
    def process(self):
        """
        Build a system image from the specified description. The
        build command combines the prepare and create commands
        """
        self.manual = Help()
        if self._help():
            return

        Privileges.check_for_root_permissions()

        abs_target_dir_path = os.path.abspath(
            self.command_args['--target-dir']
        )
        build_dir = os.sep.join([abs_target_dir_path, 'build'])
        image_root = os.sep.join([build_dir, 'image-root'])
        Path.create(build_dir)

        if not self.global_args['--logfile']:
            log.set_logfile(
                os.sep.join([abs_target_dir_path, 'build', 'image-root.log'])
            )

        self.load_xml_description(
            self.command_args['--description'], self.global_args['--kiwi-file']
        )

        build_checks = self.checks_before_command_args
        build_checks.update(
            {
                'check_target_directory_not_in_shared_cache':
                    [abs_target_dir_path],
                'check_target_dir_on_unsupported_filesystem':
                    [abs_target_dir_path]
            }
        )
        self.run_checks(build_checks)

        if self.command_args['--set-type-attr']:
            for set_type_attr in self.command_args['--set-type-attr']:
                (attribute, value) = self.attr_token(set_type_attr)
                log.info(f'--> Set <type ... {attribute}="{value}" .../>')
                try:
                    eval(
                        f'self.xml_state.build_type.set_{attribute}("{value}")'
                    )
                except AttributeError as issue:
                    log.error(f'Failed to set type attribute: {issue}')
                    return

        if self.command_args['--set-release-version']:
            release_version = self.command_args['--set-release-version']
            log.info(f'--> Set <release-version> = {release_version}')
            section_overwrite = False
            for preferences in self.xml_state.get_preferences_sections():
                section = preferences.get_release_version()
                if section:
                    section[0] = release_version
                    section_overwrite = True
                    break
            if not section_overwrite:
                preferences = self.xml_state.get_preferences_sections()[0]
                preferences.add_release_version(release_version)

        if self.command_args['--ignore-repos']:
            self.xml_state.delete_repository_sections()
        elif self.command_args['--ignore-repos-used-for-build']:
            self.xml_state.delete_repository_sections_used_for_build()

        if self.command_args['--set-repo']:
            self.xml_state.set_repository(
                *self._get_repo_parameters(
                    self.command_args['--set-repo'],
                    self.command_args['--set-repo-credentials']
                )
            )

        if self.command_args['--add-repo']:
            for add_repo, add_credentials in zip_longest(
                self.command_args['--add-repo'],
                self.command_args['--add-repo-credentials']
            ):
                self.xml_state.add_repository(
                    *self._get_repo_parameters(add_repo, add_credentials)
                )

        if self.command_args['--set-container-tag']:
            self.xml_state.set_container_config_tag(
                self.command_args['--set-container-tag']
            )

        if self.command_args['--add-container-label']:
            for add_label in self.command_args['--add-container-label']:
                try:
                    (name, value) = add_label.split('=', 1)
                    self.xml_state.add_container_config_label(name, value)
                except Exception:
                    log.warning(
                        'Container label {0} ignored. Invalid format: '
                        'expected labelname=value'.format(add_label)
                    )

        if self.command_args['--set-container-derived-from']:
            self.xml_state.set_derived_from_image_uri(
                self.command_args['--set-container-derived-from']
            )

        self.run_checks(self.checks_after_command_args)

        log.info('Preparing new root system')
        with SystemPrepare(
            self.xml_state,
            image_root,
            self.command_args['--allow-existing-root']
        ) as system:
            with system.setup_repositories(
                self.command_args['--clear-cache'],
                self.command_args[
                    '--signing-key'
                ] + self.xml_state.get_repositories_signing_keys(),
                self.global_args['--target-arch']
            ) as manager:
                system.install_bootstrap(
                    manager, self.command_args['--add-bootstrap-package']
                )

                setup = SystemSetup(
                    self.xml_state, image_root
                )
                setup.import_description()

                # call post_bootstrap.sh script if present
                setup.call_post_bootstrap_script()

                system.install_system(
                    manager
                )
                if self.command_args['--add-package']:
                    system.install_packages(
                        manager, self.command_args['--add-package']
                    )
                if self.command_args['--delete-package']:
                    system.delete_packages(
                        manager, self.command_args['--delete-package']
                    )

                profile = Profile(self.xml_state)

                defaults = Defaults()
                defaults.to_profile(profile)
                profile.create(
                    Defaults.get_profile_file(image_root)
                )

                setup.import_overlay_files()
                setup.import_image_identifier()
                setup.setup_groups()
                setup.setup_users()
                setup.setup_keyboard_map()
                setup.setup_locale()
                setup.setup_plymouth_splash()
                setup.setup_timezone()
                setup.setup_permissions()
                setup.import_files()
                setup.setup_registry_import()

                # setup permanent image repositories after cleanup
                setup.import_repositories_marked_as_imageinclude()

                # call config.sh script if present
                setup.call_config_script()

                # if configured, assign SELinux labels
                setup.setup_selinux_file_contexts()

                # handle uninstall package requests, gracefully uninstall
                # with dependency cleanup
                system.pinch_system(force=False)

                # handle delete package requests, forced uninstall without
                # any dependency resolution
                system.pinch_system(force=True)

                # create file list of packages if requested
                setup.create_system_files()

                # delete any custom rpm macros created
                system.clean_package_manager_leftovers()

        setup.call_image_script()

        log.info('Creating system image')
        self.run_checks(
            {
                'check_dracut_module_versions_compatible_to_kiwi':
                    [image_root]
            }
        )
        image_builder = ImageBuilder.new(
            self.xml_state,
            abs_target_dir_path,
            image_root,
            custom_args={
                'signing_keys': self.command_args[
                    '--signing-key'
                ] + self.xml_state.get_repositories_signing_keys(),
                'xz_options': self.runtime_config.get_xz_options()
            }
        )
        result = image_builder.create()
        result.print_results()
        result.dump(
            os.sep.join([abs_target_dir_path, 'kiwi.result'])
        )

    def _help(self):
        if self.command_args['help']:
            self.manual.show('kiwi::system::build')
        else:
            return False
        return self.manual

    def _get_repo_parameters(self, tokens, credentials):
        parameters = self.eleventuple_token(tokens)
        signing_keys_index = 6
        repo_source_index = 0
        if not parameters[signing_keys_index]:
            # make sure to pass empty list for signing_keys param
            parameters[signing_keys_index] = []
        if credentials:
            if os.path.isfile(credentials):
                credentials_data = open(credentials).readline().strip(os.linesep)
                os.unlink(credentials)
                credentials = credentials_data
            repo_source = parameters[repo_source_index]
            repo_scheme = urlparse(repo_source).scheme
            if repo_scheme:
                repo_source = repo_source.replace(f'{repo_scheme}://', '')
                repo_source = f'{repo_scheme}://{credentials}@{repo_source}'
                parameters[repo_source_index] = repo_source
        return parameters



# Copyright (c) 2024 SUSE Software Solutions Germany GmbH.  All rights reserved.
#
# This file is part of kiwi.
#
# kiwi is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# kiwi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kiwi.  If not, see <http://www.gnu.org/licenses/>
#
from typing import Dict

# project
from kiwi.system.identifier import SystemIdentifier
from kiwi.bootloader.config.base import BootLoaderConfigBase


class BootLoaderConfigCustom(BootLoaderConfigBase):
    """
    **custom bootloader configuration.**
    """
    def setup_disk_boot_images(
        self, boot_uuid, efi_uuid=None, lookup_path=None
    ) -> None:
        raise NotImplementedError

    def setup_disk_image_config(
        self, boot_uuid: str = '', root_uuid: str = '', hypervisor: str = '',
        kernel: str = '', initrd: str = '', boot_options: Dict[str, str] = {}
    ) -> None:
        raise NotImplementedError

    def setup_install_boot_images(
        self, mbrid: SystemIdentifier, lookup_path: str = ''
    ) -> None:
        raise NotImplementedError

    def setup_install_image_config(
        self, mbrid: SystemIdentifier, hypervisor: str = '',
        kernel: str = '', initrd: str = ''
    ) -> None:
        raise NotImplementedError

    def setup_live_boot_images(
        self, mbrid: SystemIdentifier, lookup_path: str = ''
    ) -> None:
        raise NotImplementedError

    def setup_live_image_config(
        self, mbrid: SystemIdentifier, hypervisor: str = '',
        kernel: str = '', initrd: str = ''
    ) -> None:
        raise NotImplementedError

    def setup_sysconfig_bootloader(self) -> None:
        raise NotImplementedError

    def write(self) -> None:
        raise NotImplementedError



# Copyright (c) 2025 SUSE LLC.  All rights reserved.
#
# This file is part of kiwi.
#
# kiwi is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# kiwi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kiwi.  If not, see <http://www.gnu.org/licenses/>
#
import os
import logging
from typing import (
    List, Dict
)

# project
from kiwi.defaults import Defaults
from kiwi.repository.base import RepositoryBase
from kiwi.path import Path

log = logging.getLogger('kiwi')


class RepositoryApk(RepositoryBase):
    """
    **Implements repository handling for apk (Alpine) package manager**
    """
    def post_init(self, custom_args: List = []) -> None:
        """
        Post initialization method

        Store custom apk arguments and create runtime configuration
        and environment

        :param list custom_args: apk arguments
        """
        self.bootstrap_repo = ''
        self.custom_args = custom_args
        self.repo_dict: Dict[str, str] = {}
        self.signing_keys: List = []
        self.repo_config = os.sep.join(
            [self.root_dir, Defaults.get_apk_repo_config()]
        )
        self.command_env = self._create_apk_get_runtime_environment()
        Path.create(os.path.dirname(self.repo_config))

    def setup_package_database_configuration(self) -> None:
        """
        Setup package database configuration

        No special database configuration required for apk
        """
        pass

    def use_default_location(self) -> None:
        """
        Call repository operations with default repository manager setup

        No special database configuration required for apk
        """
        pass

    def runtime_config(self) -> Dict:
        """
        apk runtime configuration and environment
        """
        return {
            'bootstrap_repo': self.bootstrap_repo,
            'command_env': self.command_env,
        }

    def add_repo(
        self, name: str, uri: str, repo_type: str = 'apk',
        prio: int = None, dist: str = None, components: str = None,
        user: str = None, secret: str = None, credentials_file: str = None,
        repo_gpgcheck: bool = None, pkg_gpgcheck: bool = None,
        sourcetype: str = None, customization_script: str = None,
        architectures: str = None
    ) -> None:
        """
        Add apk repository

        :param str name: repository name
        :param str uri: repository URI
        :param str repo_type: unused
        :param int prio: unused
        :param str dist: unused
        :param str components: unused
        :param str user: unused
        :param str secret: unused
        :param str credentials_file: unused
        :param bool repo_gpgcheck: unused
        :param bool pkg_gpgcheck: unused
        :param str sourcetype: unused
        :param str customization_script: unused
        :param str architectures: unused
        """
        self.repo_dict[name] = uri
        if not self.bootstrap_repo:
            # The first repo added is used for bootstrap
            self.bootstrap_repo = uri
        with open(self.repo_config, 'a') as repos:
            repos.write(f'{uri}{os.linesep}')

    def import_trusted_keys(self, signing_keys: List) -> None:
        """
        Imports trusted keys into the image

        :param list signing_keys: list of the key files to import
        """
        pass

    def delete_repo(self, name: str) -> None:
        """
        Delete apk repository

        :param str name: repository base file name
        """
        if self.repo_dict.get(name):
            self.delete_all_repos()
            del self.repo_dict[name]
            for name in self.repo_dict:
                self.add_repo(name, self.repo_dict[name])

    def delete_all_repos(self) -> None:
        """
        Delete all apk repositories
        """
        if os.path.isfile(self.repo_config):
            os.unlink(self.repo_config)

    def delete_repo_cache(self, name: str) -> None:
        """
        Delete repository cache

        No special cache cleanup required for apk

        :param str name: unused
        """
        pass

    def cleanup_unused_repos(self) -> None:
        """
        Cleanup/Delete unused repositories

        No special cleanup required for apk
        """
        pass

    def _create_apk_get_runtime_environment(self) -> Dict:
        return dict(
            os.environ, LANG='C'
        )



# Copyright (c) 2015 SUSE Linux GmbH.  All rights reserved.
#
# This file is part of kiwi.
#
# kiwi is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# kiwi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kiwi.  If not, see <http://www.gnu.org/licenses/>
#
import os
import logging
import time
import pathlib

# project
from kiwi.command import Command
from kiwi.storage.device_provider import DeviceProvider
from kiwi.utils.command_capabilities import CommandCapabilities

from kiwi.exceptions import (
    KiwiLoopSetupError,
    KiwiCommandNotFound,
    KiwiCommandError
)

log = logging.getLogger('kiwi')


class LoopDevice(DeviceProvider):
    """
    **Create and manage loop device file for block operations**

    :param string filename: loop file name to create
    :param int filesize_mbytes: size of the loop file
    :param int blocksize_bytes: blocksize used in loop driver
    """
    def __init__(
        self, filename: str, filesize_mbytes: int = None,
        blocksize_bytes: int = None
    ):
        self.node_name = ''
        if not os.path.exists(filename) and not filesize_mbytes:
            raise KiwiLoopSetupError(
                'Can not create loop file without a size'
            )
        self.filename = filename
        self.filesize_mbytes = filesize_mbytes
        self.blocksize_bytes = blocksize_bytes

    def __enter__(self):
        return self

    def get_device(self) -> str:
        """
        Device node name

        :return: device node name

        :rtype: str
        """
        return self.node_name

    def is_loop(self) -> bool:
        """
        Always True

        :return: True

        :rtype: bool
        """
        return True

    def create(self, overwrite: bool = True):
        """
        Setup a loop device of the blocksize given in the constructor
        The file to loop is created with the size specified in the
        constructor unless an existing one should not be overwritten

        :param bool overwrite: overwrite existing file to loop
        """
        if overwrite:
            qemu_img_size = format(self.filesize_mbytes) + 'M'
            Command.run(
                ['qemu-img', 'create', self.filename, qemu_img_size]
            )
        loop_options = []
        if self.blocksize_bytes and self.blocksize_bytes != 512:
            if CommandCapabilities.has_option_in_help(
                'losetup', '--sector-size', raise_on_error=False
            ):
                loop_options.append('--sector-size')
            else:
                loop_options.append('--logical-blocksize')
            loop_options.append(format(self.blocksize_bytes))
        loop_call = Command.run(
            ['losetup'] + loop_options + ['-f', '--show', self.filename]
        )
        self.node_name = loop_call.output.rstrip(os.linesep)

    def __exit__(self, exc_type, exc_value, traceback):
        if self.node_name:
            try:
                Command.run(['losetup', '-d', self.node_name])
                # loop detach is an async operation, re-use of the device
                # should not be done before the block device has been released
                loop_released = False
                sys_block_loop = pathlib.Path(
                    '/sys/devices/virtual/block/{0}/loop'.format(
                        os.path.basename(self.node_name)
                    )
                )
                for busy in range(0, 100):
                    if not sys_block_loop.is_block_device():
                        loop_released = True
                        break
                    time.sleep(0.1)
                if not loop_released:
                    raise KiwiCommandError(
                        f'Loop device {self.node_name} still attached'
                    )
            except (KiwiCommandError, KiwiCommandNotFound) as issue:
                log.error(
                    'loop cleanup on {0} failed with: {1}'.format(
                        self.node_name, issue
                    )
                )



# Copyright (c) 2022 Marcus Schäfer
#
# This file is part of kiwi.
#
# kiwi is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# kiwi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kiwi.  If not, see <http://www.gnu.org/licenses/>
#
import os
import glob
import logging
from string import Template
from contextlib import ExitStack
from typing import Dict

# project
from kiwi.path import Path
from kiwi.bootloader.template.systemd_boot import BootLoaderTemplateSystemdBoot
from kiwi.bootloader.config.bootloader_spec_base import BootLoaderSpecBase
from kiwi.boot.image.base import BootImageBase
from kiwi.mount_manager import MountManager
from kiwi.storage.loop_device import LoopDevice
from kiwi.storage.disk import Disk
from kiwi.command import Command
from kiwi.utils.os_release import OsRelease

from kiwi.exceptions import (
    KiwiTemplateError,
    KiwiBootLoaderTargetError,
    KiwiKernelLookupError
)

import kiwi.defaults as defaults

log = logging.getLogger('kiwi')


class BootLoaderSystemdBoot(BootLoaderSpecBase):
    def create_loader_image(self, target: str) -> None:
        """
        Handle EFI images

        For systemd boot EFI images are provided along with systemd.
        Thus the only action taken care of is the creation of the
        ESP path

        :param str target:
            target identifier, one of disk, live(iso) or install(iso)
        """
        self.efi_boot_path = self.create_efi_path()

    def setup_loader(self, target: str) -> None:
        """
        Setup ESP for systemd-boot using bootctl

        For disk images only, setup the ESP layout for systemd boot
        using bootctl. All data will be loaded from the ESP.

        :param str target:
            target identifier, one of disk, live(iso) or install(iso)
            Currently only the disk identifier is supported
        """
        if target != self.target.disk:
            raise KiwiBootLoaderTargetError(
                'systemd-boot is only supported with the EFI disk image target'
            )
        boot_options = self.custom_args['boot_options']
        self._mount_system(
            boot_options.get('root_device'),
            boot_options.get('boot_device'),
            boot_options.get('efi_device'),
            boot_options.get('system_volumes')
        )
        dracut_setup = self.xml_state.get_dracut_config('setup')
        self._run_bootctl(self.root_mount.mountpoint)
        if dracut_setup.uefi:
            boot_image = boot_options.get('boot_image')
            if boot_image:
                log.info('Creating UKI EFI binary')
                Command.run(
                    [
                        'mv', boot_image.create_uki(self.cmdline),
                        f'{self.root_mount.mountpoint}/boot/efi/EFI/Linux/'
                    ]
                )
        else:
            log.info('Creating loader entry')
            self.set_loader_entry(
                self.root_mount.mountpoint, self.target.disk
            )

    def set_loader_entry(self, root_dir: str, target: str) -> None:
        """
        Setup/update loader entries

        :param str target:
            target identifier, one of disk, live(iso) or install(iso)
        """
        os_release = OsRelease(root_dir)
        try:
            kernel_name = os.path.basename(
                list(glob.iglob(f'{root_dir}/lib/modules/*'))[0]
            )
        except Exception as issue:
            raise KiwiKernelLookupError(
                f'Kernel lookup in {root_dir}/lib/modules failed with {issue}'
            )
        default_entry = f'{os_release.get("ID")}-{kernel_name}.conf'
        BootLoaderSystemdBoot._write_config_file(
            BootLoaderTemplateSystemdBoot().get_entry_template(),
            root_dir + f'/boot/efi/loader/entries/{default_entry}',
            self._get_template_parameters(default_entry)
        )

    def _create_embedded_fat_efi_image(self, path: str):
        """
        Creates a EFI system partition image at the given path.
        Installs systemd-boot required EFI layout into the image
        """
        fat_image_mbsize = int(
            self.xml_state.build_type
                .get_efifatimagesize() or defaults.EFI_FAT_IMAGE_SIZE
        )
        Command.run(
            ['qemu-img', 'create', path, f'{fat_image_mbsize}M']
        )
        Command.run(
            ['sgdisk', '-n', ':1.0', '-t', '1:EF00', path]
        )
        with LoopDevice(path) as loop_provider:
            loop_provider.create(overwrite=False)
            with Disk('gpt', loop_provider) as disk:
                disk.map_partitions()
                disk.partitioner.partition_id = 1
                disk._add_to_map('efi')
                Command.run(
                    ['mkdosfs', '-n', 'BOOT', disk.partition_map['efi']]
                )
                Path.create(f'{self.root_dir}/boot/efi')
                with ExitStack() as stack:
                    efi_mount = MountManager(
                        device=disk.partition_map['efi'],
                        mountpoint=f'{self.root_dir}/boot/efi'
                    )
                    stack.push(efi_mount)
                    device_mount = MountManager(
                        device='/dev',
                        mountpoint=f'{self.root_dir}/dev'
                    )
                    stack.push(device_mount)
                    proc_mount = MountManager(
                        device='/proc',
                        mountpoint=f'{self.root_dir}/proc'
                    )
                    stack.push(proc_mount)
                    sys_mount = MountManager(
                        device='/sys',
                        mountpoint=f'{self.root_dir}/sys'
                    )
                    stack.push(sys_mount)
                    efi_mount.mount()
                    device_mount.bind_mount()
                    proc_mount.bind_mount()
                    sys_mount.bind_mount()
                    self._run_bootctl(self.root_dir)
                    self.set_loader_entry(self.root_dir, self.target.live)
                Command.run(
                    ['dd', f'if={disk.partition_map["efi"]}', f'of={path}.img']
                )

        Command.run(
            ['mv', f'{path}.img', path]
        )

    def _run_bootctl(self, root_dir: str) -> None:
        """
        Setup ESP for systemd-boot using bootctl
        """
        kernel_info = BootImageBase(
            self.xml_state, root_dir, root_dir
        ).get_boot_names()
        Command.run(
            [
                'chroot', root_dir, 'bootctl', 'install',
                '--esp-path=/boot/efi',
                '--no-variables',
                '--entry-token', 'os-id'
            ]
        )
        self._write_kernel_cmdline_file(root_dir)

        # copy kernel and initrd
        dracut_setup = self.xml_state.get_dracut_config('setup')
        if not dracut_setup.uefi:
            entry_dir = f'{root_dir}/boot/efi/loader/entries'
            os_dir = f'{root_dir}/boot/efi/os'
            Path.create(entry_dir)
            Path.create(os_dir)
            self.custom_args['kernel'] = \
                f'os/{os.path.basename(kernel_info.kernel_filename)}'
            self.custom_args['initrd'] = \
                f'os/{kernel_info.initrd_name}'
            Command.run(
                ['cp', kernel_info.kernel_filename, os_dir]
            )
            Command.run(
                ['cp', f'{root_dir}/boot/{kernel_info.initrd_name}', os_dir]
            )

        # create loader.conf
        BootLoaderSystemdBoot._write_config_file(
            BootLoaderTemplateSystemdBoot().get_loader_template(),
            root_dir + '/boot/efi/loader/loader.conf',
            self._get_template_parameters()
        )

    def _get_template_parameters(
        self, default_entry: str = 'main.conf'
    ) -> Dict[str, str]:
        return {
            'kernel_file': self.custom_args['kernel'] or 'vmlinuz',
            'initrd_file': self.custom_args['initrd'] or 'initrd',
            'boot_options': self.cmdline,
            'boot_timeout': self.timeout,
            'bootpath': self.get_boot_path(),
            'title': self.get_menu_entry_title(),
            'default_entry': default_entry
        }

    @staticmethod
    def _write_config_file(
        template: Template, filename: str, parameters: Dict[str, str]
    ) -> None:
        try:
            config_data = template.substitute(parameters)
            Path.create(os.path.dirname(filename))
            with open(filename, 'w') as config:
                config.write(config_data)
        except Exception as issue:
            raise KiwiTemplateError(
                '{0}: {1}'.format(type(issue).__name__, issue)
            )

    def _write_kernel_cmdline_file(self, root_dir: str) -> None:
        kernel_cmdline = f'{root_dir}/etc/kernel/cmdline'
        Path.create(os.path.dirname(kernel_cmdline))
        with open(kernel_cmdline, 'w') as cmdline:
            cmdline.write(self.cmdline)



from pytest import raises

from kiwi.utils.size import StringToSize

from kiwi.exceptions import KiwiSizeError


class TestStringToSize:
    def test_to_bytes(self):
        assert StringToSize.to_bytes('1m') == 1048576
        assert StringToSize.to_bytes('1M') == 1048576
        assert StringToSize.to_bytes('1g') == 1073741824
        assert StringToSize.to_bytes('1G') == 1073741824

    def test_to_bytes_wrong_format(self):
        with raises(KiwiSizeError):
            StringToSize.to_bytes('1mb')



# Copyright (c) 2015 SUSE Linux GmbH.  All rights reserved.
#
# This file is part of kiwi.
#
# kiwi is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# kiwi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kiwi.  If not, see <http://www.gnu.org/licenses/>
#
import os
import errno
import logging
from stat import ST_MODE
from typing import List

# project
from kiwi.command import Command

log = logging.getLogger('kiwi')


class DataSync:
    """
    **Sync data from a source directory to a target directory**
    """
    def __init__(self, source_dir: str, target_dir: str) -> None:
        """
        Create a new DataSync instance and initialize
        sync source and target

        :param str source_dir: source directory path name
        :param str target_dir: target directory path name
        """
        self.source_dir = source_dir
        self.target_dir = target_dir

    def sync_data(
        self, options: List[str] = [], exclude: List[str] = [],
        force_trailing_slash: bool = False
    ) -> None:
        """
        Sync data from source to target using the rsync protocol

        :param list options: rsync options
        :param list exclude: file patterns to exclude
        :param bool force_trailing_slash: add '/' to source_dir if not present

        A speciality of the rsync tool is that it behaves differently
        if the given source_dir ends with a '/' or not. If it ends
        with a slash the data structure below will be synced to the
        target_dir. If it does not end with a slash the source_dir
        and its contents are synced to the target_dir. For example

        .. code:: bash

            source
              └── some_data

            1. $ rsync -a source target

            target
              └── source
                    └── some_data

            2. $ rsync -a source/ target

            target
              └── some_data

        The parameter force_trailing_slash can be used to make
        sure rsync behaves like shown in the second case. If
        set to true a '/' is appended to the given source_dir
        if not already present
        """
        if force_trailing_slash and not self.source_dir.endswith(os.sep):
            self.source_dir += os.sep
        target_entry_permissions = None
        exclude_options = []
        rsync_options = []
        if options:
            rsync_options = options
        if not self.target_supports_extended_attributes():
            warn_me = False
            if '--xattrs' in rsync_options:
                rsync_options.remove('--xattrs')
                warn_me = True
            if '--acls' in rsync_options:
                rsync_options.remove('--acls')
                warn_me = True
            if warn_me:
                log.warning(
                    'Extended attributes not supported for target: %s',
                    self.target_dir
                )
        if exclude:
            for item in exclude:
                exclude_options.append('--exclude')
                exclude_options.append(
                    '/' + item
                )
        if os.path.exists(self.target_dir):
            target_entry_permissions = os.stat(self.target_dir)[ST_MODE]
        Command.run(
            ['rsync'] + rsync_options + exclude_options + [
                self.source_dir, self.target_dir
            ]
        )
        if target_entry_permissions:
            # rsync applies the permissions of the source directory
            # also to the target directory which is unwanted because
            # only permissions of the files and directories from the
            # source directory and its contents should be transfered
            # but not from the source directory itself. Therefore
            # the permission bits of the target directory before the
            # sync are applied back after sync to ensure they have
            # not changed
            os.chmod(self.target_dir, target_entry_permissions)

    def target_supports_extended_attributes(self) -> bool:
        """
        Check if the target directory supports extended filesystem
        attributes

        :return: True or False

        :rtype: bool
        """
        try:
            os.getxattr(self.target_dir, 'user.mime_type')
        except OSError as e:
            log.debug(
                f'Check for extended attributes on {self.target_dir} said: {e}'
            )
            if e.errno == errno.ENOTSUP:
                return False
        return True



# Copyright (c) 2015 SUSE Linux GmbH.  All rights reserved.
#
# This file is part of kiwi.
#
# kiwi is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# kiwi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kiwi.  If not, see <http://www.gnu.org/licenses/>
#
import collections
import os

# project
from kiwi.command import Command

from kiwi.exceptions import (
    KiwiContainerSetupError
)


class ContainerSetupBase:
    """
    Base class for setting up the root system to create
    a container image from for e.g docker. The methods here
    are generic to linux systems following the FHS standard
    and modern enough e.g based on systemd

    Attributes

    * :attr:`root_dir`
        root directory path name

    * :attr:`custom_args`
        dict of custom arguments
    """
    def __init__(self, root_dir, custom_args=None):
        if not os.path.exists(root_dir):
            raise KiwiContainerSetupError(
                f'Container root directory {root_dir} does not exist'
            )
        self.root_dir = root_dir
        self.custom_args = custom_args or {}
        if 'container_name' not in self.custom_args:
            self.custom_args['container_name'] = 'system-container'
        self.post_init(custom_args)

    def post_init(self, custom_args):
        """
        Post initialization method

        Implementation in specialized container setup class

        :param list custom_args: unused
        """
        pass

    def setup(self):
        """
        Setup container metadata

        Implementation in specialized bootloader class required
        """
        raise NotImplementedError

    def deactivate_bootloader_setup(self):
        """
        Container bootloader setup

        Tell the system there is no bootloader configuration
        it needs to care for. A container does not boot
        """
        bootloader_setup = self.root_dir + '/etc/sysconfig/bootloader'
        if os.path.exists(bootloader_setup):
            self._update_config(
                bootloader_setup,
                {
                    'LOADER_LOCATION': 'LOADER_LOCATION="none"',
                    'LOADER_TYPE': 'LOADER_TYPE="none"'
                }
            )

    def deactivate_root_filesystem_check(self):
        """
        Container filesystem check setup

        The root filesystem of a container could be an overlay
        or a mapped device. In any case it should not be checked
        for consistency as this is should be done by the container
        infrastructure
        """
        boot_setup = self.root_dir + '/etc/sysconfig/boot'
        if os.path.exists(boot_setup):
            self._update_config(
                boot_setup,
                {
                    'ROOTFS_BLKDEV': 'ROOTFS_BLKDEV="/dev/null"'
                }
            )

    def deactivate_systemd_service(self, name):
        """
        Container system services setup

        Init systems among others also controls services which
        starts at boot time. A container does not really boot.
        Thus some services needs to be deactivated

        :param string name: systemd service name
        """
        service_file = self.root_dir + '/usr/lib/systemd/system/' + name
        if os.path.exists(service_file):
            try:
                Command.run(
                    ['ln', '-s', '-f', '/dev/null', service_file]
                )
            except Exception as e:
                raise KiwiContainerSetupError(
                    f'Failed to deactivate service {name}: {format(e)}'
                )

    def setup_root_console(self):
        """
        Container console setup

        /dev/console should be allowed to login by root
        """
        securetty = self.root_dir + '/etc/securetty'
        if not os.path.exists(securetty):
            with open(securetty, 'w'):
                pass
        self._update_config(
            securetty,
            {
                'console': 'console'
            }
        )

    def get_container_name(self):
        """
        Container name

        :return: name
        :rtype: str
        """
        return self.custom_args['container_name']

    def _update_config(self, filename, update_record):
        data = []
        with open(filename, 'r') as config:
            data = config.read().rsplit('\n')

        sorted_record = collections.OrderedDict(
            sorted(update_record.items())
        )
        for current_value, new_value in list(sorted_record.items()):
            entry_found = False
            for index in range(0, len(data)):
                line = data[index]
                if line.startswith(current_value):
                    entry_found = True
                    data[index] = new_value
            if not entry_found:
                data.append(new_value)

        with open(filename, 'w') as config:
            config.write('%s\n' % '\n'.join(data))



# Copyright (c) 2015 SUSE Linux GmbH.  All rights reserved.
#
# This file is part of kiwi.
#
# kiwi is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# kiwi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kiwi.  If not, see <http://www.gnu.org/licenses/>
#
import logging
import importlib
from typing import List
from abc import (
    ABCMeta,
    abstractmethod
)

# project
from kiwi.system.uri import Uri
from kiwi.exceptions import KiwiRootImportError

log = logging.getLogger('kiwi')


class RootImport(metaclass=ABCMeta):
    """
    Root import factory

    Attibutes

    * :attr:`root_dir`
        root directory path name

    * :attr:`image_uri`
        an instance of :class:`Uri` to an image containing a the root system

    * :attr:`image_type`
        type of the image to import
    """
    @abstractmethod
    def __init__(self) -> None:
        return None  # pragma: no cover

    @staticmethod
    def new(root_dir: str, image_uri: List[Uri], image_type: str):
        name_map = {
            'docker': 'OCI',
            'oci': 'OCI'
        }
        log.info(
            'Importing root from a {0} image type'.format(image_type)
        )
        (custom_args, module_namespace) = \
            RootImport._custom_args_for_root_import(image_type)
        try:
            rootimport = importlib.import_module(
                'kiwi.system.root_import.{0}'.format(module_namespace)
            )
            module_name = 'RootImport{0}'.format(name_map[module_namespace])
            return rootimport.__dict__[module_name](
                root_dir, image_uri, custom_args
            )
        except Exception as issue:
            raise KiwiRootImportError(
                'Support to import {0} images not implemented: {1}'.format(
                    image_type, issue
                )
            )

    @staticmethod
    def _custom_args_for_root_import(image_type: str):
        if image_type == 'docker':
            custom_args = {'archive_transport': 'docker-archive'}
        else:
            custom_args = {'archive_transport': 'oci-archive'}
        return [custom_args, 'oci']



import pytest

from .conftest import (
    LEAP_15_5,
    TUMBLEWEED,
)


CONTAINER_IMAGES = [TUMBLEWEED]


def test_does_nothing_when_product_correct(auto_container_per_test):
    previous_contents = auto_container_per_test.connection.file(
        "/etc/products.d"
    ).listdir()
    auto_container_per_test.connection.run_expect(
        [0], ". /bin/functions.sh && suseSetupProduct"
    )
    assert auto_container_per_test.connection.file("/etc/products.d").listdir() == previous_contents


@pytest.mark.parametrize(
    "container_per_test,product_name",
    (
        (TUMBLEWEED, "openSUSE.prod"),
        (LEAP_15_5, "Leap.prod"),
    ),
    indirect=["container_per_test"],
)
def test_sets_baseproduct_from_etc_os_release(container_per_test, product_name):
    assert not container_per_test.connection.file("/etc/SuSE-brand").exists

    container_per_test.connection.run_expect([0], "rm /etc/products.d/baseproduct")
    container_per_test.connection.run_expect(
        [0], ". /bin/functions.sh && suseSetupProduct"
    )

    assert container_per_test.connection.file("/etc/products.d/baseproduct").exists
    assert container_per_test.connection.file("/etc/products.d/baseproduct").is_symlink
    assert container_per_test.connection.file("/etc/products.d/baseproduct").linked_to == "/etc/products.d/" + product_name


def test_sets_baseproduct_with_weird_os_release(auto_container_per_test):
    assert not auto_container_per_test.connection.file(
        "/etc/SuSE-brand"
    ).exists
    auto_container_per_test.connection.run_expect(
        [0], "rm /etc/products.d/baseproduct"
    )
    auto_container_per_test.connection.run_expect(
        [0],
        """cat <<EOF > /etc/os-release
NAME=openSUSE Tumbleweed
ID="opensuse-tumbleweed"
ID_LIKE="opensuse suse"
EOF
""",
    )

    auto_container_per_test.connection.run_expect(
        [0], ". /bin/functions.sh && suseSetupProduct"
    )
    assert auto_container_per_test.connection.file(
        "/etc/products.d/baseproduct"
    ).exists
    assert auto_container_per_test.connection.file(
        "/etc/products.d/baseproduct"
    ).is_symlink
    assert (
        auto_container_per_test.connection.file(
            "/etc/products.d/baseproduct"
        ).linked_to == "/etc/products.d/openSUSE.prod"
    )


def test_sets_baseproduct_from_prod_files(auto_container_per_test):
    auto_container_per_test.connection.run_expect(
        [0],
        """rm /etc/products.d/baseproduct
rm /etc/products.d/openSUSE.prod
touch /etc/products.d/10.prod
touch /etc/products.d/20.prod
""",
    )

    auto_container_per_test.connection.run_expect(
        [0], ". /bin/functions.sh && suseSetupProduct"
    )
    assert auto_container_per_test.connection.file(
        "/etc/products.d/baseproduct"
    ).exists
    assert auto_container_per_test.connection.file(
        "/etc/products.d/baseproduct"
    ).is_symlink
    assert (
        auto_container_per_test.connection.file(
            "/etc/products.d/baseproduct"
        ).linked_to == "/etc/products.d/20.prod"
    )



from kiwi.builder.template.container_import import BuilderTemplateSystemdUnit


class TestBuilderTemplateSystemdUnit:
    def setup_method(self):
        self.template = BuilderTemplateSystemdUnit()

    def test_get_container_import_template_with_container_snap_service(self):
        container_files = ['/var/tmp/kiwi_containers/test-app_v1.0']
        load_commands = [['/usr/bin/container-snap', 'load', '-i', '/var/tmp/kiwi_containers/test-app_v1.0']]
        after_services = ['container-snap.service']

        template = self.template.get_container_import_template(
            container_files, load_commands, after_services
        )

        unit_content = template.substitute()

        expected_unit = """# kiwi generated unit file
[Unit]
Description=Import Local Container(s)
ConditionPathExists=/var/tmp/kiwi_containers/test-app_v1.0
After=container-snap.service
[Service]
Type=oneshot
ExecStart=/usr/bin/container-snap load -i /var/tmp/kiwi_containers/test-app_v1.0
ExecStartPost=/bin/rm -f /var/tmp/kiwi_containers/test-app_v1.0
[Install]
WantedBy=multi-user.target
"""

        assert unit_content == expected_unit

    def test_get_container_import_template_with_multiple_services(self):
        container_files = [
            '/var/tmp/kiwi_containers/app1_latest',
            '/var/tmp/kiwi_containers/app2_v1.0'
        ]
        load_commands = [
            ['/usr/bin/docker', 'load', '-i', '/var/tmp/kiwi_containers/app1_latest'],
            ['/usr/bin/container-snap', 'load', '-i', '/var/tmp/kiwi_containers/app2_v1.0']
        ]
        after_services = ['docker.service', 'container-snap.service']

        template = self.template.get_container_import_template(
            container_files, load_commands, after_services
        )

        unit_content = template.substitute()

        expected_unit = """# kiwi generated unit file
[Unit]
Description=Import Local Container(s)
ConditionPathExists=/var/tmp/kiwi_containers/app1_latest
ConditionPathExists=/var/tmp/kiwi_containers/app2_v1.0
After=docker.service container-snap.service
[Service]
Type=oneshot
ExecStart=/usr/bin/docker load -i /var/tmp/kiwi_containers/app1_latest
ExecStart=/usr/bin/container-snap load -i /var/tmp/kiwi_containers/app2_v1.0
ExecStartPost=/bin/rm -f /var/tmp/kiwi_containers/app1_latest
ExecStartPost=/bin/rm -f /var/tmp/kiwi_containers/app2_v1.0
[Install]
WantedBy=multi-user.target
"""

        assert unit_content == expected_unit

    def test_get_container_import_template_no_after_services(self):
        container_files = ['/var/tmp/kiwi_containers/test-app_latest']
        load_commands = [['/usr/bin/podman', 'load', '-i', '/var/tmp/kiwi_containers/test-app_latest']]
        after_services = []

        template = self.template.get_container_import_template(
            container_files, load_commands, after_services
        )

        unit_content = template.substitute()

        expected_unit = """# kiwi generated unit file
[Unit]
Description=Import Local Container(s)
ConditionPathExists=/var/tmp/kiwi_containers/test-app_latest
[Service]
Type=oneshot
ExecStart=/usr/bin/podman load -i /var/tmp/kiwi_containers/test-app_latest
ExecStartPost=/bin/rm -f /var/tmp/kiwi_containers/test-app_latest
[Install]
WantedBy=multi-user.target
"""

        assert unit_content == expected_unit

    def test_get_container_import_template_empty_inputs(self):
        template = self.template.get_container_import_template([], [], [])

        unit_content = template.substitute()

        expected_unit = """# kiwi generated unit file
[Unit]
Description=Import Local Container(s)
[Service]
Type=oneshot
[Install]
WantedBy=multi-user.target
"""

        assert unit_content == expected_unit



import io
from collections import namedtuple
from unittest.mock import (
    patch, Mock, MagicMock
)
from pytest import raises

from kiwi.boot.image.base import BootImageBase
from kiwi.defaults import Defaults

from kiwi.exceptions import (
    KiwiTargetDirectoryNotFound,
    KiwiConfigFileNotFound,
    KiwiDiskBootImageError
)


class TestBootImageBase:
    @patch('kiwi.boot.image.base.os.path.exists')
    def setup(self, mock_exists):
        Defaults.set_platform_name('x86_64')
        self.boot_names_type = namedtuple(
            'boot_names_type', [
                'kernel_name', 'initrd_name',
                'kernel_version', 'kernel_filename'
            ]
        )
        self.kernel = Mock()
        kernel_info = Mock
        kernel_info.name = 'kernel_name'
        kernel_info.version = 'kernel_version'
        kernel_info.filename = 'kernel_filename'
        self.kernel.get_kernel.return_value = kernel_info
        self.boot_xml_state = Mock()
        self.xml_state = Mock()
        self.xml_state.get_initrd_system = Mock(
            return_value='dracut'
        )
        self.xml_state.xml_data.get_name = Mock(
            return_value='some-image'
        )
        self.xml_state.get_image_version = Mock(
            return_value='1.2.3'
        )
        self.xml_state.build_type.get_boot = Mock(
            return_value='oemboot/suse-13.2'
        )
        mock_exists.return_value = True
        self.boot_image = BootImageBase(
            self.xml_state, 'some-target-dir', 'system-directory'
        )

    @patch('kiwi.boot.image.base.os.path.exists')
    def setup_method(self, cls, mock_exists):
        self.setup()

    def test_boot_image_raises(self):
        with raises(KiwiTargetDirectoryNotFound):
            BootImageBase(
                self.xml_state, 'target-dir-does-not-exist', 'some-root-dir'
            )

    def test_prepare(self):
        with raises(NotImplementedError):
            self.boot_image.prepare()

    def test_create_initrd(self):
        with raises(NotImplementedError):
            self.boot_image.create_initrd()

    def test_create_uki(self):
        with raises(NotImplementedError):
            self.boot_image.create_uki('some_cmdline')

    @patch('os.listdir')
    def test_is_prepared(self, mock_listdir):
        mock_listdir.return_value = []
        assert self.boot_image.is_prepared() is False
        mock_listdir.return_value = ['a', 'b', 'c']
        assert self.boot_image.is_prepared() is True

    @patch('kiwi.boot.image.base.XMLState.copy_strip_sections')
    def test_import_system_description_elements(self, mock_strip):
        self.boot_image.import_system_description_elements()
        assert self.xml_state.copy_displayname.called
        assert self.xml_state.copy_name.called
        assert self.xml_state.copy_repository_sections.called
        assert self.xml_state.copy_drivers_sections.called
        assert mock_strip.called
        assert self.xml_state.copy_strip_sections.called
        assert self.xml_state.copy_preferences_subsections.called
        assert self.xml_state.copy_bootincluded_packages.called
        assert self.xml_state.copy_bootincluded_archives.called
        assert self.xml_state.copy_bootdelete_packages.called
        assert self.xml_state.copy_build_type_attributes.called
        assert self.xml_state.copy_systemdisk_section.called
        assert self.xml_state.copy_machine_section.called
        assert self.xml_state.copy_oemconfig_section.called

    def test_get_boot_description_directory(self):
        assert self.boot_image.get_boot_description_directory() == \
            '/usr/share/kiwi/custom_boot/oemboot/suse-13.2'

    @patch('kiwi.boot.image.base.BootImageBase.get_boot_description_directory')
    def test_load_boot_xml_description(self, mock_boot_dir):
        mock_boot_dir.return_value = None
        with raises(KiwiConfigFileNotFound):
            self.boot_image.load_boot_xml_description()

    @patch('kiwi.boot.image.base.Kernel')
    def test_get_boot_names_raises(self, mock_Kernel):
        kernel = Mock()
        mock_Kernel.return_value = kernel
        kernel.get_kernel.return_value = None
        with raises(KiwiDiskBootImageError):
            self.boot_image.get_boot_names()

    @patch('kiwi.boot.image.base.Kernel')
    def test_get_boot_names_with_initrd_system_set_to_none(self, mock_Kernel):
        kernel = Mock()
        mock_Kernel.return_value = kernel
        kernel.get_kernel.return_value = None
        self.xml_state.get_initrd_system.return_value = 'none'
        boot_names = self.boot_image.get_boot_names()
        assert boot_names.kernel_name == 'none'
        assert boot_names.initrd_name == 'none'
        assert boot_names.kernel_version == 'none'

    @patch('kiwi.boot.image.base.Kernel')
    @patch('kiwi.boot.image.base.Path.which')
    @patch('kiwi.boot.image.base.log.warning')
    @patch('glob.iglob')
    def test_get_boot_names_default(
        self, mock_iglob, mock_warning, mock_Path_which, mock_Kernel
    ):
        mock_iglob.return_value = []
        mock_Path_which.return_value = None
        mock_Kernel.return_value = self.kernel
        self.xml_state.get_initrd_system.return_value = 'kiwi'
        assert self.boot_image.get_boot_names() == self.boot_names_type(
            kernel_name='kernel_name',
            initrd_name='initrd-kernel_version',
            kernel_version='kernel_version',
            kernel_filename='kernel_filename'
        )
        self.xml_state.get_initrd_system.return_value = 'dracut'
        assert self.boot_image.get_boot_names() == self.boot_names_type(
            kernel_name='kernel_name',
            initrd_name='initramfs-kernel_version.img',
            kernel_version='kernel_version',
            kernel_filename='kernel_filename'
        )

    @patch('kiwi.boot.image.base.Kernel')
    @patch('kiwi.boot.image.base.Path.which')
    @patch('kiwi.boot.image.base.log.warning')
    @patch('glob.iglob')
    @patch('os.path.islink')
    def test_get_boot_names_from_file(
        self, mock_islink, mock_iglob, mock_warning, mock_Path_which,
        mock_Kernel
    ):
        mock_islink.return_value = False
        mock_iglob.return_value = [
            '/boot/initrd.img-kernel_version'
        ]
        mock_Path_which.return_value = None
        mock_Kernel.return_value = self.kernel
        self.xml_state.get_initrd_system.return_value = 'dracut'
        assert self.boot_image.get_boot_names() == self.boot_names_type(
            kernel_name='kernel_name',
            initrd_name='initrd.img-kernel_version',
            kernel_version='kernel_version',
            kernel_filename='kernel_filename'
        )

    @patch('kiwi.boot.image.base.Kernel')
    @patch('kiwi.boot.image.base.Path.which')
    @patch('kiwi.boot.image.base.log.warning')
    @patch('glob.iglob')
    def test_get_boot_names_from_dracut(
        self, mock_iglob, mock_warning, mock_Path_which, mock_Kernel
    ):
        mock_iglob.return_value = []
        mock_Path_which.return_value = 'dracut'
        mock_Kernel.return_value = self.kernel
        self.xml_state.get_initrd_system.return_value = 'dracut'
        with patch('builtins.open', create=True) as mock_open:
            mock_open.return_value = MagicMock(spec=io.IOBase)
            file_handle = mock_open.return_value.__enter__.return_value
            file_handle.read.return_value = 'outfile="/boot/initrd-$kernel"'
            assert self.boot_image.get_boot_names() == self.boot_names_type(
                kernel_name='kernel_name',
                initrd_name='initrd-kernel_version',
                kernel_version='kernel_version',
                kernel_filename='kernel_filename'
            )
            file_handle.read.return_value = 'outfile="/boot/initrd-${kernel}"'
            assert self.boot_image.get_boot_names() == self.boot_names_type(
                kernel_name='kernel_name',
                initrd_name='initrd-kernel_version',
                kernel_version='kernel_version',
                kernel_filename='kernel_filename'
            )

    def test_noop_methods(self):
        self.boot_image.include_module('module')
        self.boot_image.omit_module('module')
        self.boot_image.set_static_modules(['module'])
        self.boot_image.include_driver('driver')
        self.boot_image.omit_driver('driver')
        self.boot_image.write_system_config_file({'config_key': 'value'})
        self.boot_image.cleanup()

    def test_has_initrd_support(self):
        assert self.boot_image.has_initrd_support() is False

    def test_add_argument(self):
        assert self.boot_image.add_argument('some') is None



from unittest.mock import patch

import unittest.mock as mock

from kiwi.defaults import Defaults
from kiwi.filesystem.erofs import FileSystemEroFs


class TestFileSystemEroFs:
    @patch('os.path.exists')
    def setup(self, mock_exists):
        mock_exists.return_value = True
        self.erofs = FileSystemEroFs(
            mock.Mock(), 'root_dir',
            custom_args={'compression': 'zstd,level=21'}
        )

    @patch('os.path.exists')
    def setup_method(self, cls, mock_exists):
        self.setup()

    @patch('kiwi.filesystem.erofs.Command.run')
    def test_create_on_file(self, mock_command):
        Defaults.set_platform_name('x86_64')
        self.erofs.create_on_file('myimage', 'label')
        mock_command.assert_called_once_with(
            [
                'mkfs.erofs', '-z', 'zstd,level=21',
                '-L', 'label', 'myimage', 'root_dir'
            ]
        )

    @patch('kiwi.filesystem.erofs.Command.run')
    def test_create_on_file_exclude_data(self, mock_command):
        Defaults.set_platform_name('x86_64')
        self.erofs.create_on_file('myimage', 'label', ['foo'])
        mock_command.assert_called_once_with(
            [
                'mkfs.erofs', '-z', 'zstd,level=21',
                '-L', 'label', '--exclude-regex=^foo$',
                'myimage', 'root_dir'
            ]
        )



import logging
from unittest.mock import (
    patch, Mock
)
from pytest import (
    raises, fixture
)

from kiwi.partitioner import Partitioner

from kiwi.exceptions import KiwiPartitionerSetupError


class TestPartitioner:
    @fixture(autouse=True)
    def inject_fixtures(self, caplog):
        self._caplog = caplog

    def test_partitioner_not_implemented(self):
        with raises(KiwiPartitionerSetupError):
            Partitioner.new('foo', Mock())

    def test_partitioner_for_arch_not_implemented(self):
        with raises(KiwiPartitionerSetupError):
            Partitioner.new('foo', Mock())

    @patch('kiwi.partitioner.gpt.PartitionerGpt')
    def test_partitioner_gpt(self, mock_gpt):
        storage_provider = Mock()
        Partitioner.new('gpt', storage_provider)
        mock_gpt.assert_called_once_with(storage_provider, None, False)

    @patch('kiwi.partitioner.msdos.PartitionerMsDos')
    def test_partitioner_msdos(self, mock_dos):
        storage_provider = Mock()
        Partitioner.new('msdos', storage_provider)
        mock_dos.assert_called_once_with(storage_provider, None, False)

    @patch('kiwi.partitioner.dasd.PartitionerDasd')
    def test_partitioner_dasd(self, mock_dasd):
        storage_provider = Mock()
        Partitioner.new('dasd', storage_provider)
        mock_dasd.assert_called_once_with(storage_provider, None, False)

    @patch('kiwi.partitioner.dasd.PartitionerDasd')
    def test_partitioner_dasd_with_custom_start_sector(self, mock_dasd):
        storage_provider = Mock()
        with self._caplog.at_level(logging.WARNING):
            Partitioner.new('dasd', storage_provider, 4096)
            mock_dasd.assert_called_once_with(storage_provider, None, False)



import logging
from pytest import fixture
from unittest.mock import (
    patch, mock_open
)

import unittest.mock as mock

from kiwi.partitioner.dasd import PartitionerDasd


class TestPartitionerDasd:
    @fixture(autouse=True)
    def inject_fixtures(self, caplog):
        self._caplog = caplog

    @patch('kiwi.partitioner.dasd.Command.run')
    @patch('kiwi.partitioner.dasd.Temporary.new_file')
    def setup(self, mock_temp, mock_command):
        self.tempfile = mock.Mock()
        self.tempfile.name = 'tempfile'

        mock_temp.return_value = self.tempfile

        disk_provider = mock.Mock()
        disk_provider.get_device = mock.Mock(
            return_value='/dev/loop0'
        )

        self.partitioner = PartitionerDasd(disk_provider)

    @patch('kiwi.partitioner.dasd.Command.run')
    @patch('kiwi.partitioner.dasd.Temporary.new_file')
    def setup_method(self, cls, mock_temp, mock_command):
        self.setup()

    @patch('kiwi.partitioner.dasd.Command.run')
    @patch('kiwi.partitioner.dasd.Temporary.new_file')
    def test_create(self, mock_temp, mock_command):
        mock_command.side_effect = Exception
        mock_temp.return_value = self.tempfile

        m_open = mock_open()
        with patch('builtins.open', m_open, create=True):
            with self._caplog.at_level(logging.DEBUG):
                self.partitioner.create('name', '100', 't.linux', ['f.active'])

        m_open.return_value.write.assert_called_once_with(
            'n\np\n\n+100M\nw\nq\n'
        )
        mock_command.assert_called_once_with(
            ['bash', '-c', 'cat tempfile | fdasd -f /dev/loop0']
        )

    @patch('kiwi.partitioner.dasd.Command.run')
    @patch('kiwi.partitioner.dasd.Temporary.new_file')
    def test_create_all_free(self, mock_temp, mock_command):
        mock_temp.return_value = self.tempfile

        m_open = mock_open()
        with patch('builtins.open', m_open, create=True):
            self.partitioner.create('name', 'all_free', 't.linux')

        m_open.return_value.write.assert_called_once_with(
            'n\np\n\n\nw\nq\n'
        )

    def test_resize_table(self):
        self.partitioner.resize_table()
